﻿using System;
using System.IO;
using System.Text.RegularExpressions;
using System.Threading;

/* (c)2017 - S.co.tt
 * License:  I dunno, some kind of open-source attribution thing.
 *           Leave this comment block in place and add your own credit,
 *           and also let me know.  Just search "The Google" for "ScottDotDot"
 *           to find a way to contact me.  Some kind of social media, 
 *           probably.  Or scott@s.co.tt if social media isn't a thing
 *           anymore by the time you read this.
 */

namespace ZeroFiller
{
    class Program
    {
        static void Main(string[] args)
        {
            string sDrive = "";
            string sFileName = "\\ZeroFiller.blob";
            Int64 lFreeSpace = 11;
            string sFreeSpaceQualifier = "%";
            long lBlockSizeMiB = 100;
            bool bDeleteOnComplete = true;
            bool bDeleteOnly = false;


            if (args.Length == 0 || (args.Length == 1 && (args[0].ToLower().Contains("help") || args[0].ToLower().Contains("?"))))
            {
                Console.WriteLine("Fills a disk up with a binary file consisting of all zeroes,");
                Console.WriteLine("  for the purpose of making the disk image compress well");
                Console.WriteLine("  during backups or migrations.");
                Console.WriteLine("");
                Console.WriteLine("Usage: ZeroFiller.exe X: [options]");
                Console.WriteLine("Options:");
                Console.WriteLine("    /fn:xxxxx  - FileName, including optional path.");
                Console.WriteLine("                 (DEFAULT @ \\ZeroFiller.blob)");
                Console.WriteLine("    /fs:xxxM   - FreeSpace to leave in MiB.");
                Console.WriteLine("    /fs:xx%    - FresSpace as percent of disk size.");
                Console.WriteLine("                 (DEFAULT option @ 11%)");
                Console.WriteLine("    /bs:xxxx   - Block size in MiB. (DEFAULT @ 100MiB)");
                Console.WriteLine("    /nd        - NoDelete, won't delete file after.");
                Console.WriteLine("                 This is a bad idea; Mainly for debug.");
                Console.WriteLine("    /do        - DeleteOnly.  Only deletes the blob file.");
                Console.WriteLine("                 Good to run periodically in case it gets interrupted.");
                Console.WriteLine("                 Overrides the /nd flag if it's set.");
                Environment.Exit(0);
            }

            Regex r = null;
            if (args.Length >= 1)
            {
                r = new Regex("^[a-zA-Z]:$");
                if (r.IsMatch(args[0]) && Directory.Exists(args[0]))
                    sDrive = args[0];
                else
                {
                    Console.WriteLine("ERROR: The first argument must be in the format X:");
                    Console.WriteLine("       Oh, and that disk must exist (natch).");
                    Environment.Exit(1);
                }
            }

            for (int i=1; i<args.Length; i++)
            {
                string sArg = args[i];

                if (sArg.StartsWith("/fn:"))
                {
                    string sFile = sArg.Substring(4);
                    try
                    {
                        if (!sFile.StartsWith("\\"))
                            sFile = "\\" + sFile;
                        
                        FileInfo fi = new FileInfo(sDrive + sFile);
                        if (fi.Exists)
                        {
                            Console.WriteLine("WARNING: File exists: {0}", sFile);
                        }
                        if (!fi.Directory.Exists)
                        {
                            Console.WriteLine("ERROR: Path does not exist: {0}", fi.DirectoryName);
                            Environment.Exit(1);
                        }
                        sFileName = sFile;
                    } catch (Exception ex) {
                        Console.WriteLine("ERROR: Exception with arg /fn:");
                        Console.WriteLine(ex.Message);
                        Environment.Exit(1);
                    }
                } else if (sArg.StartsWith("/fs:"))
                {
                    r = new Regex("^[0-9]+(M|%)$");

                    string sFreeSpace = sArg.Substring(4);

                    if (!r.IsMatch(sFreeSpace))
                    {
                        Console.WriteLine("ERROR: Argument /fs: was not in a valid format.");
                        Environment.Exit(1);
                    }
                    try
                    {
                        lFreeSpace = Convert.ToInt64(sFreeSpace.Substring(0, sFreeSpace.Length - 1));
                        sFreeSpaceQualifier = sFreeSpace.Substring(sFreeSpace.Length - 1);
                    } catch (Exception ex) {
                        Console.WriteLine("ERROR: Exception coercing free space to long int (probably).");
                        Environment.Exit(1);
                    }
                } else  if (sArg.StartsWith("/bs:"))
                {
                    try
                    {
                        lBlockSizeMiB = Convert.ToInt64(sArg.Substring(4));
                    } catch (Exception ex) {
                        Console.WriteLine("ERROR: Block size could not be coerced to a long int.");
                        Environment.Exit(1);
                    }
                } else if (sArg == "/nd")
                {
                    bDeleteOnComplete = false;
                } else if (sArg == "/do")
                {
                    bDeleteOnly = true;
                } else {
                    Console.WriteLine("ERROR: Invalid argument: {0}", sArg);
                    Environment.Exit(1);
                }
            }

            Console.WriteLine("[{0}] Starting.", DateTime.Now.ToString("yyyy-MM-dd HH:mm:ss"));

            string sFileNameFull = sDrive + sFileName;

            if (bDeleteOnly)
            {
                Console.WriteLine("WARNING: /do flag is set.  Only deleting {0} .", sFileNameFull);
                if (File.Exists(sFileNameFull))
                    DeleteBlobFile(sFileNameFull);
                else
                    Console.WriteLine("File did not exist.  Exiting.");
                Environment.Exit(0);
            }
            
            Int64 lBlocksWritten = 0;
            Int64 lExistingFileSize = 0;

            Console.WriteLine("Processing file {0} ..", sFileNameFull);

            if (File.Exists(sFileNameFull))
            {
                FileInfo fiExisting = new FileInfo(sFileNameFull);
                lExistingFileSize = Convert.ToInt64(Math.Round((double)fiExisting.Length / 1048576d));
                Console.WriteLine("File exists at {0}MiB, utilizing it.. ", fiExisting.Length / 1048576); 
            }

            Stream sw = new FileStream(sFileNameFull, FileMode.Append);
            BinaryWriter bw = new BinaryWriter(sw);

            try
            {
                DriveInfo di = new DriveInfo(sDrive);

                //Setting a max amount of blocks means that this won't frack things up if the free space constantly increases (e.g. a large file is deleted and then re-created).
                //It also means that the write loop won't get carried away to infinity.
                Int64 lMaxBlocks = 0;

                //Math.
                //I'm not Stephen Hawking, so this might be fucked.  I mean, it works, but it's idiotically implemented.  PARENTHESES!!!!  ORDER OF OPERATIONS!!!
                if (sFreeSpaceQualifier == "%")
                    lMaxBlocks = (Int64)(((((double)di.TotalFreeSpace / (double)di.TotalSize - ((double)lFreeSpace / 100d)) * di.TotalSize) / 1048576d) - lExistingFileSize) / lBlockSizeMiB;
                if (sFreeSpaceQualifier == "M")
                    lMaxBlocks = (di.TotalFreeSpace - lFreeSpace * 1048576) / (lBlockSizeMiB * 1048576);

                Console.WriteLine("Block size: {0}MiB", lBlockSizeMiB);
                Console.WriteLine("Free space now: {0}MiB ({1}% free)", Math.Round((double)di.TotalFreeSpace / 1048576d, 0), Math.Round((100 * (double)di.TotalFreeSpace / di.TotalSize), 1));
                Console.WriteLine("Writing until {0}{1} free on disk {2}", lFreeSpace, sFreeSpaceQualifier, sDrive);
                Console.WriteLine("Max of {0} blocks ({1}MiB) to be written", lMaxBlocks, lMaxBlocks * lBlockSizeMiB);
                Console.WriteLine("Writing file..");

                DateTime dtStart = DateTime.Now;

                for (lBlocksWritten = 0; lBlocksWritten < lMaxBlocks; lBlocksWritten++)
                {
                    //I know I could just throw the bFull-equivalent-maths into the IF statement and break right there, but this made it easier for debugging.
                    //Also I think I was gonna do more with that flag, but then didn't.  Ah, well.
                    bool bFull = false;
                    if (sFreeSpaceQualifier == "%")
                        bFull = 100d * ((double)di.TotalFreeSpace / (double)di.TotalSize) <= (double)lFreeSpace;
                    if (sFreeSpaceQualifier == "M")
                        bFull = (double)di.TotalFreeSpace / 1048576d <= (double)lFreeSpace;
                    if (bFull)
                        break;

                    bw.Write(new Byte[lBlockSizeMiB * 1048576]);
                    bw.Flush();
                    sw.Flush();
                    bw.Close();
                    sw.Close();
                    bw.Dispose();
                    sw.Dispose();

                    //This seems really inefficient, but it's the only way to get NTFS to recognize the actual file size & free space.
                    //Just flushing it like the toilet it is?  That doesn't do the trick.
                    sw = new FileStream(sFileNameFull, FileMode.Append);
                    bw = new BinaryWriter(sw);

                    // It looks like the GC doesn't clean fast enough.  Don't want to hog memory on a server.  (Better to hog CPU! ;] )
                    if (lBlocksWritten * lBlockSizeMiB >= 2048)
                    {
                        GC.Collect();
                        GC.WaitForPendingFinalizers();
                    }
                }

                DateTime dtEnd = DateTime.Now;

                Console.WriteLine("Disk free threshold reached.");
                Console.WriteLine("Wrote {0} blocks ({1}MiB)", lBlocksWritten, lBlocksWritten * lBlockSizeMiB);
                FileInfo fiDone = new FileInfo(sFileNameFull);
                Console.WriteLine("Actual file size: {0}MiB", fiDone.Length / 1048576);
                Console.WriteLine("Took {0} minutes at {1}MB/s", Math.Round((double)dtEnd.Subtract(dtStart).TotalSeconds / 60d, 1), Math.Round(((double)lBlockSizeMiB * (double)lBlocksWritten) / (double)dtEnd.Subtract(dtStart).TotalSeconds, 1));
                Console.WriteLine("Current free space: {0}MiB ({1}%)", Math.Round((double)di.TotalFreeSpace / 1048576d, 0), Math.Round((100 * (double)di.TotalFreeSpace / di.TotalSize), 1));
            } catch (IOException ex)
            {
                Console.WriteLine("ERROR: IOException while writing file..");
                Console.WriteLine(ex.Message);
                Environment.Exit(1);
            } catch (Exception ex)
            {
                Console.WriteLine("ERROR: Exception while writing file..");
                Console.WriteLine(ex.Message);
                Environment.Exit(1);
            } finally
            {
                bw.Flush();
                sw.Flush();
                bw.Close();
                sw.Close();
                bw.Dispose();
                sw.Dispose();
            }

            if (bDeleteOnComplete)
            {
                Console.WriteLine("Deleting blob file..");
                DeleteBlobFile(sFileNameFull);
            }
            else
                Console.WriteLine("WARNING: Not deleting blob file per /nd flag.");

            Console.WriteLine("[{0}] DONE.", DateTime.Now.ToString("yyyy-MM-dd HH:mm:ss"));

            //Console.ReadLine();
        }

        public static void DeleteBlobFile(string sFile)
        {
            //I had a problem wherein the file lock didn't get released right away on one system.  Dunno why.  You'd think a flush/close/dispose should damn well be enough.
            //So I just ghetto-fabulously retry the delete every 5 seconds for a max of 5 retries.  It works.  Leave it, or you'll be scrathing your head one day while
            //NTFS is fucking you over.  Now that I think about it, maybe it has to do with THUMBS.DB for absolutely NO REASON.  Have you ever tried to delete a file on
            //a network share using Win7, only to be cockblocked by the thing that generates THUMBS.DB?  No?  Well look it up.  They went like 10 years without fixing that crap.
            int incre = 0;
            while (File.Exists(sFile))
            {
                Console.WriteLine("Deleting {0} ..", sFile);
                try
                {
                    File.Delete(sFile);
                    return;
                }
                catch (IOException ex)
                {
                    Console.WriteLine("ERROR: Could not delete {0} ..", sFile);
                    Console.WriteLine("  This is the #{0} try.", incre);
                    Console.WriteLine("IOException: {0}", ex.Message);
                }
                catch (Exception ex)
                {
                    Console.WriteLine("ERROR: Could not delete {0} ..", sFile);
                    Console.WriteLine("  This is the #{0} try.", incre);
                    Console.WriteLine("Exception: {0}", ex.Message);
                }
                incre++;
                if (incre > 5)
                {
                    Console.WriteLine("ERROR: Could not delete file after 5 retries.");
                    Console.WriteLine("   {0}", sFile);
                    Environment.Exit(1);
                }
                
                Thread.Sleep(5000);
            }
        }
    }
}
